% =================================================================
:- pred constr(bool).
:- mode constr(in).
:- ignore constr/1.

% =================================================================
% Program. ascending sorting
:- pred bubblesort(list(int),list(int)).
:- mode bubblesort(in,out).

bubblesort(L,L) :- swap(L,L1,B), constr(~B), eqlist(L,L1). 
bubblesort(L,SL) :- swap(L,L1,B), constr(B),           
                    bubblesort(L1,SL).
                    
:- pred eqlist(list(int),list(int)).
:- mode eqlist(in,in).
eqlist(L,L).

:- pred swap(list(int),list(int),bool). % bool is true if a swap has been done
:- mode swap(in,out,out).

swap([],[],B) :- constr(~B).
swap([X],[X],B) :- constr(~B).
swap([X,Y|Ys],[Y|Zs],B) :- 
         constr(B & (X > Y)),
         cons(X,Ys,XYs),
         swap(XYs,Zs,B).
swap([X,Y|Ys],[X|Zs],B) :-
        constr(X =< Y),
        cons(Y,Ys,YYs),
        swap(YYs,Zs,B).

:- pred cons(int,list(int),list(int)).
:- mode cons(in,in,out).
cons(H,T,[H|T]).

% -------------------------------------
% catamorphism
:- pred count(int,list(int),int).
:- mode count(in,in,out).
:- cata count/3-2.

count(X,[],Res) :- constr( Res=0 ).
count(X,[Y|Ys],Res) :-
    constr( Res=ite(X=Y,(ResT+1),ResT) ),
    count(X,Ys,ResT).

% =================================================================
% verification.
% Property:
:- pred ff1.
ff1 :-
   constr(~(CL = CS)),
   count(X,L,CL), count(X,S,CS),
   bubblesort(L,S).

:- pred ff2.
ff2 :-
   constr( ~(CLA = CLB) ),
   count(X,LA,CLA), count(X,LB,CLB),
   swap(LA,LB,ToF).

:- pred ff3.
ff3 :-
   constr( ~(CHT = ite(X=H, (CT+1), CT)) ),
   count(X,T,CT), count(X,HT,CHT),
   cons(H,T,HT).

% =================================================================
:- query ff1/0.
:- query ff2/0.
:- query ff3/0.
% =================================================================


