% =================================================================
:- pred constr(bool).
:- mode constr(in).
:- ignore constr/1.

% =================================================================
% Program. ascending sorting
:- pred bubblesort(list(int),list(int)).
:- mode bubblesort(in,out).

bubblesort(L,L) :- swap(L,L1,B), constr(~B), eqlist(L,L1). 
bubblesort(L,SL) :- swap(L,L1,B), constr(B),           
                    bubblesort(L1,SL).
                    
:- pred eqlist(list(int),list(int)).
:- mode eqlist(in,in).
eqlist(L,L).

:- pred swap(list(int),list(int),bool).   % bool is true if a swap has been done
:- mode swap(in,out,out).

swap([],[],B) :- constr(~B).
swap([X],[X],B) :- constr(~B).
swap([X,Y|Ys],[Y|Zs],B) :- 
         constr(B & (X > Y)),
         cons(X,Ys,XYs),
         swap(XYs,Zs,B).
swap([X,Y|Ys],[X|Zs],B) :-
        constr(X =< Y),
        cons(Y,Ys,YYs),
        swap(YYs,Zs,B).

:- pred cons(int,list(int),list(int)).
:- mode cons(in,in,out).
cons(H,T,[H|T]).

% =================================================================
% catamorphism
:- pred length(list(int),int).
:- mode length(in,out).
:- cata length/2-1.

length([],L) :- constr((L=0)).
length([H|T],L) :-
  constr( (L=LT+1) ),
  length(T,LT).

% =================================================================
% Verification.
% Property:
:- pred ff1.
ff1 :-
   constr(~(CL = CS)),
   length(L,CL), length(S,CS),
   bubblesort(L,S).

:- pred ff2.
ff2 :-
   constr( ~(CLA = CLB) ),
   length(LA,CLA), length(LB,CLB),
   swap(LA,LB,ToF).

:- pred ff3.
ff3 :-
   constr( ~(LHT = LT+1) ),
   length(T,LT), length(HT,LHT),
   cons(H,T,HT).

% =================================================================
:- query ff1/0.
:- query ff2/0.
:- query ff3/0.
% =================================================================


