% =================================================================
:- pred constr(bool).
:- mode constr(in).
:- ignore constr/1.

% =================================================================
% Program. ascending sorting
:- pred insertionsort(list(int),list(int)).
:- mode insertionsort(in,out).

insertionsort([],[]).
insertionsort([X|Xs],S) :- ord_ins(X,TS,S),
         insertionsort(Xs,TS).

:- pred ord_ins(int,list(int),list(int)).
:- mode ord_ins(in,in,out).

ord_ins(A,[],[A]).
ord_ins(A,[B|Xs],[A,B|Xs]) :- A=<B.
ord_ins(A,[B|Xs],[B|AXs]) :- A>B,
     ord_ins(A,Xs,AXs).

% =================================================================
% catamorphisms
%% checking the weakly ascending order in a list
:- pred is_asorted(list(int),bool).
:- mode is_asorted(in,out).
:- cata is_asorted/2-1.

is_asorted([],Res) :- constr(Res).
is_asorted([H|T],Res) :-
    hd(T,IsDefT,HdT),
    is_asorted(T,ResT),
    constr(Res=(IsDefT => (H=<HdT & ResT))).

% ---
:- pred hd(list(int),bool,int).
:- mode hd(in,out,out).
:- cata hd/3-1.

hd([],IsDef,Hd) :- constr( ~IsDef & (Hd=0) ).    % Any int value for Hd, instead of 0.
hd([H|T],IsDef,Hd) :- constr( IsDef & (Hd=H) ).


% =================================================================
% verification.
% Property:
:- pred ff1.
ff1:-
   constr(~B),
   is_asorted(S,B),
   insertionsort(L,S).

:- pred ff2.
ff2 :-
   constr( ~( SL = SXL ) ),
   is_asorted(L,SL), is_asorted(XL,SXL),
   ord_ins(X,L,XL).

% =================================================================
:- query ff1/0.
:- query ff2/0.
% ================================================================= 


