% ================================================================= 
:- pred constr(bool). 
:- mode constr(in).
:- ignore constr/1.
% ================================================================= 
% Program. ascending sorting  (ascending because we take the minimum)  2 =< 5 =< 9 =< ...
:- pred selectionsort(list(int),list(int)).
:- mode selectionsort(in,out).
selectionsort([],[]).
selectionsort([H|T],[Min|T1]) :-
 cons(H,T,HT),        
 listMin(HT,IsDef,Min),              % HT is not empty. IsDef is true.     
 delete(Min,HT,L1),                  % Taking the minimum away from list HT.
 selectionsort(L1,T1).               

:- pred delete(int,list(int),list(int)).
:- mode delete(in,in,in).
%delete(X,[],[]).                    % delete is not total
delete(X,[Y|T],T) :- constr(X=Y).    % taking X away from [Y|T]. 
delete(X,[Y|T],[Y|TD]) :-            % Deleting exactly one copy of X only. 
  constr(~(X=Y)),                    % The list is not ordered. 
  delete(X,T,TD).                    % The element X need not be the minimum.

:- pred cons(int,list(int),list(int)).
:- mode cons(in,in,out).             
cons(H,T,[H|T]).
% ====================================================
% catamorphism

:- pred listMin(list(int),bool,int).     
:- mode listMin(in,out,out).
:- cata listMin/3-1.

listMin([],IsDef,A) :- constr( ((~IsDef) & (A=0)) ).
listMin([H|T],IsDef,Min) :-
         constr(IsDef & (Min = ite(IsDefT, ite(H<MinT,H,MinT), H)) ),
         listMin(T,IsDefT,MinT).
         
% ---
:- pred len(list(int),int).
:- mode len(in,out).
:- cata len/2-1.

len([],L) :- L=0.
len([H|T],L) :-
      L=(LT+1),
      len(T,LT).

% ====================================================
% verification. 
% Property:
:- pred ff1.
ff1 :- 
   constr( ~(LL=LS)), 
   len(L,LL), len(S,LS),  
   selectionsort(L,S).

:- pred ff2.
ff2 :- 
   constr( ~((X=MinL) => ( LLD = (LL-1) )) ),
   len(L,LL), len(LD,LLD), listMin(L,IsDefMinL,MinL),
   delete(X,L,LD).

:- pred ff3.
ff3 :- 
  constr( ~(LHT = (LT+1)) ),
  len(T,LT), len(HT,LHT),
  cons(H,T,HT).

% ====================================================
:- query ff1/0.
:- query ff2/0.
:- query ff3/0.
% ====================================================

