% ================================================================= 
:- pred constr(bool). 
:- mode constr(in).
:- ignore constr/1.
% ====================================================
% Program. sorting in ascending order.E.g., 3 =< 5 =< 5 =< 8 =< ...
:- pred selectionsort(list(int),list(int)).
:- mode selectionsort(in,out).

selectionsort([],[]).  
selectionsort([H|T],[Min|T1]) :-
  cons(H,T,HT),        
  listMin(HT,IsDef,Min),             % HT is not empty. IsDef is true.     
  delete(Min,HT,L1),                 % Taking the minimum away from list HT.
  selectionsort(L1,T1).   

:- pred delete(int,list(int),list(int)).
:- mode delete(in,in,in).
%delete(X,[],[]).                    % delete in not total
delete(X,[Y|T],T) :- constr(X=Y).    % Taking X away from [Y|T]. 
delete(X,[Y|T],[Y|TD]) :-            % Deleting exactly one copy of X only. 
  constr(~(X=Y)),                    % The list is not ordered. 
  delete(X,T,TD).                    % The element X need not be the minimum.

:- pred cons(int,list(int),list(int)).
:- mode cons(in,in,out).             
cons(H,T,[K|T]) :- constr(K=H-1). % correct: cons(H,T,[H|T]).

% ====================================================
% catamorphisms
:- pred listMin(list(int),bool,int).
:- mode listMin(in,out,out).                  
:- cata listMin/3-1.

listMin([],IsDef,A) :- constr( ((~IsDef) & (A=0)) ). % 0 can be any int.
listMin([H|T],IsDef,Min) :-
         constr( IsDef & (Min = ite(IsDefT, ite(H<MinT,H,MinT), H )) ),
         listMin(T,IsDefT,MinT).

% ================================================================= 
% verification. 
% Property:
:- pred ff1.
ff1 :-
   constr(~( (IsDefL = IsDefS) & (IsDefL => (MinL=MinS)) )),
   listMin(L,IsDefL,MinL), listMin(S,IsDefS,MinS),
   selectionsort(L,S).

% Contract on delete (for any X, not necessarily the minimum of L): 
:- pred ff2.
ff2 :- 
    constr(~(  (   IsDefL  & 
               (  (IsDefLD & (X>MinL)) => (MinL = MinLD) ))  & 
               (( ~IsDefLD) => (MinL=X))
             )),
    listMin(L,IsDefL,MinL), listMin(LD,IsDefLD,MinLD),
    delete(X,L,LD).

:- pred ff3.
ff3 :- 
   constr(~( IsDefHT & ( MinHT=ite(IsDefT,ite(H<MinT,H,MinT),H) ) )),
   listMin(T,IsDefT,MinT), listMin(HT,IsDefHT,MinHT),
   cons(H,T,HT).
% =================================================================
:- query ff1/0.
:- query ff2/0.
:- query ff3/0.
% =================================================================

